#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
"""
audio.cd (Compact Disc)
=======================

**Author:**

* Dirk Alders <sudo-dirk@mount-mockery.de>

**Description:**

    This Module is a submodule of :mod:`audio` and a long description can be placed here.

**Submodules:**

* :mod:`mmod.module.sub1`
* :class:`mmod.module.sub2`
* :func:`mmod.module.sub2`

**Unittest:**

        See also the :download:`unittest <../../audio/_testresults_/unittest.pdf>` documentation.
"""
import report

import CDDB
import DiscID
import os
import subprocess


class compact_disc(list, report.logit):
    """
    Class for audio compact discs

    :param str device: The device (e.g. /dev/sr0)
    :param str track_comment: An optional comment for the tracks (see :py:class:track)
    :param str basepath: An optional basepath for the tracks (see :py:class:track)
    :param logger: An optional logger instance
    """
    LOG_PREFIX = 'CD:'

    def __init__(self, device, track_comment='', basepath=None, logger=None):
        list.__init__(self)
        self._device = device
        self._track_comment = track_comment
        self._basepath = basepath
        self.disc_id = None
        self.read_disc_data(logger=logger)

    def read_disc_data(self, logger=None):
        """Method to read the disc content creating a list of :py:class:`track`s including cddb data if available

        :param logger: An optional logger instance
        """
        cdrom = DiscID.open(self._device)
        try:
            old_did = self.disc_id
            self.disc_id = DiscID.disc_id(cdrom)
            self.logit(logger, report.logging.INFO, 'DISC-ID = %s', self.disc_id)
        except:
            self.logit(logger, report.logging.WARNING, 'DISC-ID could not be detected!')
        finally:
            cdrom.close()
        if old_did != self.disc_id:
            self._process_cddb_data(logger=logger)

    def _process_cddb_data(self, logger=None):
        def decode(string):
            for i in ['utf-8', 'cp1252']:
                try:
                    return string.decode(i)
                except UnicodeEncodeError:
                    pass
                except UnicodeDecodeError:
                    pass
            return string
        # TODO: default values on fail
        list.__init__(self)
        try:
            query_info = CDDB.query(self.disc_id)[1]
            if type(query_info) == list:
                query_info = query_info[0]
            if type(query_info) == dict:
                cddb_data = CDDB.read(query_info['category'], query_info['disc_id'])[1]
            else:
                cddb_data = None
        except:
            self.logit(logger, report.logging.WARNING, 'Error reading CD-Information (CDDB), check network connection')
            cddb_data = None
        finally:
            number_of_tracks = self.disc_id[1]
            if cddb_data is not None:
                genre = cddb_data['DGENRE']
                artist = [txt.strip() for txt in cddb_data['DTITLE'].split('/')][0]
                album = '/'.join([txt.strip() for txt in cddb_data['DTITLE'].split('/')][1:])
                self.logit(logger, report.logging.INFO, 'Got data from CDDB: %s - %s', artist, album)
                try:
                    year = int(cddb_data['DYEAR'])
                except ValueError:
                    year = 0
                self.logit(logger, report.logging.DEBUG, 'Number of Tracks = %d', number_of_tracks)
                self.logit(logger, report.logging.DEBUG, 'Artist = %s', artist)
                self.logit(logger, report.logging.DEBUG, 'Album = %s', album)
                self.logit(logger, report.logging.DEBUG, 'Genre = %s', genre)
                self.logit(logger, report.logging.DEBUG, 'Year = %d', year)
                for no in range(0, number_of_tracks):
                    ti = cddb_data['TTITLE%d' % no]
                    #if no == 7:
                    #    ti = ti.decode('iso8859_10')
                    track_data = {track.KEY_ARTIST: decode(artist),
                                  track.KEY_ALBUM: decode(album),
                                  track.KEY_GENRE: decode(genre),
                                  track.KEY_YEAR: year,
                                  track.KEY_TITLE: decode(ti),
                                  track.KEY_TRACK: no + 1,
                                  track.KEY_COMMENT: decode(self._track_comment),
                                  track.KEY_FILE_PATH: self._basepath}
                    self.logit(logger, report.logging.DEBUG, 'Track%(' + track.KEY_TRACK + ')d = %(' + track.KEY_TITLE + ')s', track_data)
                    self.append(track(track_data))
            else:
                self.logit(logger, report.logging.WARNING, 'No Data from CDDB')
                for no in range(0, number_of_tracks):
                    track_data = {track.KEY_ARTIST: 'Unknown Artist',
                                  track.KEY_ALBUM: 'Unknown Album',
                                  track.KEY_GENRE: 'Unknown Genre',
                                  track.KEY_YEAR: 0,
                                  track.KEY_TITLE: 'Unknown Title %d' % (no + 1),
                                  track.KEY_TRACK: no + 1,
                                  track.KEY_COMMENT: self._track_comment,
                                  track.KEY_FILE_PATH: self._basepath}
                    self.append(track(track_data))

    def __str__(self):
        rv = "Number of Tracks: %d\n" % len(self)
        if self._number_of_tracks > 0:
            rv += "Artist: %s\n" % self[0].get(track.KEY_ARTIST)
            rv += "Album: %s\n" % self[0].get(track.KEY_ALBUM)
            rv += "Genre: %s\n" % self[0].get(track.KEY_GENRE)
            rv += "Year: %d\n" % self[0].get(track.KEY_YEAR)
            rv += "Comment: %d\n" % self[0].get(track.KEY_COMMENT)
            for no in range(0, self._number_of_tracks):
                rv += '    %3d: %s\n' % (self[no][track.KEY_TRACK], self[no][track.KEY_TITLE])
        return rv


def rip(device, track_no, file_path, logger_callback=None, progress_callback=None, *args, **kwargs):
    """Method to rip a track.

    :param str device: The name of the device (e.g. /dev/sr0)
    :param progress_callback: A callback for reporting the rip progress

    .. note:: The progress_callback will be executed with track instance and progress as first two argumnets, followed by *args and **kwargs.
    """
    cdp_secstarttag = 'sector'
    cdp_secendtag = '('
    cdp_starttag = '[wrote] @'
    cdp_endtag = '\n'

    cmd = 'cdparanoia --stderr-progress -d %s %d %s' % (device, track_no, file_path)
    if logger_callback is not None:
        logger_callback('Executing - %s', cmd)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    line = ''
    startsec = None
    endsec = None
    startwrote = None
    progress = None
    while True:
        ch = p.stdout.read(1)
        if len(ch) == 0:
            break
        line += ch
        if line.find(cdp_secstarttag) > -1:
            line = line[line.find(cdp_secstarttag):]
            if line.find(cdp_secstarttag) > -1 and line.find(cdp_secendtag) > line.find(cdp_secstarttag):
                if startsec is None:
                    startsec = float(line[line.find(cdp_secstarttag) + len(cdp_secstarttag):line.find(cdp_secendtag)])
                    if logger_callback is not None:
                        logger_callback('secstart: %s', str(startsec))
                elif endsec is None:
                    endsec = float(line[line.find(cdp_secstarttag) + len(cdp_secstarttag):line.find(cdp_secendtag)])
                    if logger_callback is not None:
                        logger_callback('secend: %s', str(endsec))
                line = line[line.find(cdp_secendtag) + len(cdp_secendtag):]
        elif line.find(cdp_starttag) > -1:
            line = line[line.find(cdp_starttag):]
            if line.find(cdp_starttag) > -1 and line.find(cdp_endtag) > line.find(cdp_starttag):
                if startwrote is None:
                    startwrote = float(line[line.find(cdp_starttag) + len(cdp_starttag):line.find(cdp_endtag)])
                    if logger_callback is not None:
                        logger_callback('wrotestart: %s', str(startwrote))
                else:
                    thiswrote = float(line[line.find(cdp_starttag) + len(cdp_starttag):line.find(cdp_endtag)])
                    endwrote = (endsec - startsec) * 1176 + startwrote
                    this_progress = (thiswrote - startwrote) / (endwrote - startwrote)
                    if progress != this_progress:
                        progress = this_progress
                        if logger_callback is not None:
                            logger_callback('Progress callback executed if given: %f', progress)
                        if progress_callback is not None:
                            progress_callback(track_no, progress, *args, **kwargs)
                line = line[line.find(cdp_endtag) + len(cdp_endtag):]
    p.wait()
    if progress != 1:
        if logger_callback is not None:
            logger_callback('Progress callback executed if given: %f', 1)
        if progress_callback is not None:
            progress_callback(track_no, 1, *args, **kwargs)
    # TODO: Error detection required (e.g. path for wav file does not exist, incomplete rip, ...)
    return True


class track(dict, report.logit):
    """
    Class to hold track information and generate useful information out of that data. See also parent :py:class:`audio.title`

    .. note:: Can be initialised like a :py:class:`dict` with all keys :py:const:`audio.KEY_*`

    :raises: TypeError: If an argument is not the correct type.
    """
    LOG_PREFIX = 'Track:'

    KEY_ARTIST = 'artist'
    """Key for the Artist name"""
    KEY_TITLE = 'title'
    """Key for the title"""
    KEY_YEAR = 'year'
    """Key for the album year"""
    KEY_ALBUM = 'album'
    """Key for the album name"""
    KEY_TRACK = 'track_no'
    """Key for the track number"""
    KEY_GENRE = 'genre'
    """Key for the genre"""
    KEY_COMMENT = 'comment'
    """Key for the comment"""
    KEY_FILE_PATH = 'file_path'
    """Key for the filepath"""
    KEY_WAV_FILE_PATH = 'wav_file_path'
    """NEW"""
    DATATYPES = {KEY_ALBUM: [str, unicode],
                 KEY_ARTIST: [str, unicode],
                 KEY_COMMENT: [str, unicode],
                 KEY_GENRE: [str, unicode],
                 KEY_TITLE: [str, unicode],
                 KEY_TRACK: [int],
                 KEY_YEAR: [int],
                 KEY_FILE_PATH: [str, unicode],
                 KEY_WAV_FILE_PATH: [str, unicode]}
    NAME_FORMATTER = '%(' + KEY_TRACK + ')02d_%(' + KEY_TITLE + ')s'
    PATH_FORMATTER = ['%(' + KEY_ARTIST + ')s',
                      '%(' + KEY_YEAR + ')04d' + '_%(' + KEY_ALBUM + ')s']

    """
    Class to hold title information

    .. note:: Can be initialised like a :py:class:`dict` with the keys :py:const:`audio.KEY_*` except of :py:const:`KEY_FILE_PATH`

    :raises: TypeError: If an argument is not the correct type.
    """
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        for key in self.DATATYPES:
            if key in self and type(self[key]) not in self.DATATYPES[key]:
                raise TypeError

    def __str__(self):
        rv = ''
        keys = self.DATATYPES.keys()
        keys.sort()
        for key in keys:
            rv += "%14s %s\n" % (key + ':', str(self.get(key, 'None')))
        return rv

    def _str_filter_(self, s):
        return s.lower().replace(' ', '_').replace('/', '_').replace('\\', '_').replace("'", '_').replace('"', '_').replace('&', '_').replace('(', '_').replace(')', '_')

    def filename(self, ext='mp3', formatter=None):
        """Method to create a filename out of the track information.

        .. note:: The filename is generated out of the classes :py:var`NAME_FORMATTER` and the stored track information
        """
        formatter = formatter or self.NAME_FORMATTER
        return self._str_filter_(formatter % self) + '.%s' % ext

    def filepath(self, ext='mp3', path_formatter=None, name_formatter=None):
        """Method to create a filepath out of the track information.

        .. note:: The filepath is generated out of the classes :py:var`NAME_FORMATTER`, :py:var`PATH_FORMATTER` and the stored track information
        """
        if path_formatter is None:
            path_formatter = path_formatter or self.PATH_FORMATTER
        p = self[self.KEY_FILE_PATH] or '.'
        for part in path_formatter:
            p = os.path.join(p, self._str_filter_(part % self))
        return os.path.join(p, self.filename(ext, formatter=name_formatter))

    def rip(self, device, name_formatter=None, path_formatter=None, logger=None, *args, **kwargs):
        l = report.logit_to_central_logger(logger, self.LOG_PREFIX)
        l.logit(report.logging.INFO, 'Ripping Track %d', self.get(self.KEY_TRACK))
        wav_filename = self.filepath('wav', path_formatter=path_formatter, name_formatter=name_formatter)
        if rip(device, self[self.KEY_TRACK], wav_filename, l.logit_debug, *args, **kwargs):
            self[self.KEY_WAV_FILE_PATH] = wav_filename

    def encode(self, name_formatter=None, path_formatter=None, logger=None, *args, **kwargs):
        from mp3 import encode
        l = report.logit_to_central_logger(logger, self.LOG_PREFIX)
        l.logit(report.logging.INFO, 'Encoding Track %d', self.get(self.KEY_TRACK))
        encode(self[self.KEY_WAV_FILE_PATH], self.filepath('mp3', path_formatter, name_formatter), self, l.logit_debug, *args, **kwargs)
