#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
"""
audio.mp3 (MP3 Audio-Module)
============================

**Author:**

* Dirk Alders <sudo-dirk@mount-mockery.de>

**Description:**

    This Module is a submodule of :mod:`audio` and a long description can be placed here.

**Submodules:**

* :mod:`mmod.module.sub1`
* :class:`mmod.module.sub2`
* :func:`mmod.module.sub2`

**Unittest:**

        See also the :download:`unittest <../../audio/_testresults_/unittest.pdf>` documentation.
"""
import caching
import fstools
import task
import multimedia

from cd import track
import subprocess
import os
import time

FULL_DEBUG = False


LAMEARGS = '-V 0 -F -h -b 160'


def encode(wav_filename, mp3_filename, track_data, logger_callback=None, progress_callback=None, *args, **kwargs):
    # TODO: LOGGING!
    lame_starttag = '('
    lame_endtag = '%)'

    id3_data = {track.KEY_ARTIST: 'unknown artist',
                track.KEY_TITLE: 'unknown title',
                track.KEY_YEAR: 0,
                track.KEY_ALBUM: 'unknown album',
                track.KEY_TRACK: 0,
                track.KEY_GENRE: 'unknown',
                track.KEY_COMMENT: ''}
    for key in id3_data:
        id3_data[key] = track_data.get(key, id3_data[key])
    id3_args_template = '--ta "%%(%s)s" --tt "%%(%s)s" --ty "%%(%s)d" --tl "%%(%s)s" --tn "%%(%s)d" --tg "%%(%s)s" --tc "%%(%s)s"' % (track.KEY_ARTIST, track.KEY_TITLE, track.KEY_YEAR, track.KEY_ALBUM, track.KEY_TRACK, track.KEY_GENRE, track.KEY_COMMENT)
    cmd = ' '.join(['lame', LAMEARGS, id3_args_template % id3_data, wav_filename, mp3_filename])
    if logger_callback is not None:
        logger_callback('Executing - %s', cmd)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    line = ''
    progress = None
    while True:
        ch = p.stdout.read(1)
        if len(ch) == 0:
            break
        line += ch
        if line.find(lame_starttag) > -1 and line.find(lame_endtag) > -1:
            try:
                this_progress = float(line[line.find(lame_starttag) + len(lame_starttag):line.find(lame_endtag)]) / 100.
                if progress != this_progress:
                    progress = this_progress
                    if logger_callback is not None:
                        logger_callback('Progress callback executed if given: %f', progress)
                    if progress_callback is not None:
                        progress_callback(progress, *args, **kwargs)
            except ValueError:
                pass
            line = ''
    p.wait()
    if progress != 1:
        if logger_callback is not None:
            logger_callback('Progress callback executed if given: %f', 1)
        if progress_callback is not None:
            progress_callback(1, *args, **kwargs)


class play(object):
    """
        SEE: http://www.fifi.org/doc/mpg321/README.remote
    """
    STATUS_DICT = {0: 'stop', 1: 'pause', 2: 'play', 3: 'end'}

    def __init__(self, gain=75, track_completed_callback=None, logger=None):
        self.logger = logger or logging.getLogger(self.__class__.__name__)
        self._gain = gain
        self._track_completed_callback = track_completed_callback
        self.__init_vars__()
        self.dt = task.delayed(0, self._start)
        self.dt.run()
        self._write = None

    def start(self, filename, logger=None):
        logit = logger or self.logger
        logit.info('Loading and starting play-back of %s', filename)
        while self._write is None:
            time.sleep(0.1)
        self._cmd(u'L ' + filename, logit)
        self._status = 2

    def play_pause(self, logger=None):
        logit = logger or self.logger
        if self._status in [1, 2]:
            if self._status == 1:
                logit.info('Resuming play-back')
            else:
                logit.info('Pausing play-back')
            self._cmd('P', logit)
        else:
            logit.info('Pause request ignored, no song is playing')

    def jump(self, tm, logger=None):
        logit = logger or self.logger
        logit.info('Jumping to %.2fs', tm)
        max_frames = self._current_frame + self._frames_left
        max_time = self._current_time + self._time_left
        frame = int(max_frames * tm / max_time)
        if frame >= 0 and frame <= max_frames:
            self._cmd('J %d' % frame, logit)

    def set_gain(self, gain, logger=None):
        logit = logger or self.logger
        if gain >= 0 and gain <= 100:
            logit.info('Setting gain to %d%%', gain)
            self._cmd('G %d' % gain, logit)
            self._gain = gain
        else:
            logit.warning('Volume out of range (0-100): %d', gain)

    def get_gain(self):
        return self._gain

    def stop(self, logger=None):
        logit = logger or self.logger
        logit.info('Stopping play-back')
        self._cmd('S', logit)

    def close(self, logger=None):
        logit = logger or self.logger
        logit.info('Closing play-back')
        self._cmd('Q', logit)
        self._write = None
        self.dt.stop()

    def join(self):
        self.dt.join()

    def _start(self):
        cmd = ['/usr/bin/mpg321', '-g', '%d' % self._gain, '-KR', 'player']
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self._write = p.stdin.write
        line = ''
        while True:
            ch = p.stdout.read(1)
            if len(ch) == 0:
                break
            if ch != '\n':
                line += ch
            else:
                self._process(line)
                line = ''
                if self._status == 3 and self._track_completed_callback is not None:
                    self.logger.debug('Executing Callback after track ended...')
                    self._track_completed_callback(self)
        p.wait()

    def __init_vars__(self):
        self._current_frame = 0
        self._frames_left = 0
        self._current_time = 0
        self._time_left = 0
        self._status = 0
        self._vers = 0
        self._layer = 0
        self._sample_rate = 0
        self._mode = ''
        self._bytes_per_frame = 0
        self._channels = 0
        self._copyright = False
        self._crc_protected = False
        self._emphasis = 0
        self._bitrate = 0
        self._extension = 0

    def _process(self, line):
        if FULL_DEBUG:
            self.logger.debug('mpg321: %s', line)
        data = line.split(' ')
        if data[0] == '@F' and len(data) == 5:
            self._current_frame = int(data[1])
            self._frames_left = int(data[2])
            self._current_time = float(data[3])
            self._time_left = float(data[4])
        elif data[0] == '@P' and len(data) == 2:
            self._status = int(data[1])
        elif data[0] == '@S' and len(data) == 13:
            self._vers = float(data[1])
            self._layer = int(data[2])
            self._sample_rate = int(data[3])
            self._mode = data[4] + '(%s)' % data[5]
            self._bytes_per_frame = int(data[6])
            self._channels = int(data[7])
            self._copyright = bool(data[8])
            self._crc_protected = bool(data[9])
            self._emphasis = int(data[10])
            self._bitrate = int(data[11])
            self._extension = int(data[12])

    def _cmd(self, txt, logit):
        if self._write is not None:
            self._write(txt + '\n')
        else:
            logit.warning('Commad request before or after program execution')

    def __str__(self):
        return """current_frame = %d
frames_left = %d
current_time = %.2f
time_left = %.2f
status = %s
vers = %d
layer = %d
sample_rate = %d
mode = %s
bytes_per_frame = %d
channels = %d
copyright = %s
crc_protected = %s
emphasis = %d
bitrate = %d
extension = %d

""" % (self._current_frame, self._frames_left, self._current_time, self._time_left, self.STATUS_DICT[self._status], self._vers, self._layer, self._sample_rate, self._mode, self._bytes_per_frame, self._channels, self._copyright, self._crc_protected, self._emphasis, self._bitrate, self._extension)

    def __del__(self):
        self.close()


class audio_info(multimedia.base_info):
    LOG_PREFIX = 'AudInfo:'

    """Class to hold and handle information of an audio file. See also parent class :py:class:`multimedia.base_info`.

    :param str filename: Name of the audio file

    **Example:**

    .. code-block:: python

        a = audio_info(os.path.join(basepath, 'example', 'music.mp3'))
        for key in a.keys():
            print key, type(a.get(key)), a.get(key)

    Will result to the following output:

    .. code-block:: text

        album <type 'str'> A/B
        title <type 'str'> No Good
        track <type 'int'> 1
        artist <type 'str'> Kaleo
        genre <type 'str'> Rock
        year <type 'int'> 2016
        duration <type 'float'> 236.094694
    """
    DATA_VERSION_NUMBER = 0.2

    ALBUM = 'album'
    ARTIST = 'artist'
    BITRATE = 'bitrate'
    DURATION = 'duration'
    GENRE = 'genre'
    TITLE = 'title'
    TRACK = 'track'
    YEAR = 'year'
    TAG_TYPES = {ALBUM: unicode,
                 ARTIST: unicode,
                 BITRATE: int,
                 DURATION: float,
                 GENRE: unicode,
                 TITLE: unicode,
                 TRACK: int,
                 YEAR: int}

    def data_version(self):
        """
        Returns the data creation method version.

        :returns: version number
        :rtype: float
        """
        return self.DATA_VERSION_NUMBER

    def uid(self):
        """
        Returns the uid of the source file.

        :returns: unique id
        :rtype: str
        """
        return fstools.uid(self.filename)

    def _ffprobe_command(self):
        return ['ffprobe', '-v', 'quiet', '-show_format', '-show_streams', self.filename]

    def _avprobe_command(self):
        return ['avprobe', '-v', 'quiet', '-show_format', '-show_streams', self.filename]

    def _get_info(self, logger=None):
        TAG_TRANSLATION = {'TAG:album': self.ALBUM,
                           'TAG:artist': self.ARTIST,
                           'bit_rate': self.BITRATE,
                           'duration': self.DURATION,
                           'TAG:genre': self.GENRE,
                           'TAG:title': self.TITLE,
                           'TAG:track': self.TRACK,
                           'TAG:date': self.YEAR}
        self._info = dict()
        try:
            try:
                ffprobe_txt = subprocess.check_output(self._avprobe_command())
            except OSError:
                ffprobe_txt = subprocess.check_output(self._ffprobe_command())
        except subprocess.CalledProcessError:
            self.logit_error(logger, "Error processing %s", self.filename)
            ffprobe_txt = ''
        for line in ffprobe_txt.splitlines():
            try:
                key, val = [snippet.strip() for snippet in line.split('=')]
            except ValueError:
                continue
            else:
                if key in TAG_TRANSLATION and key not in self._info:
                    # some special interpretation
                    if key == 'TAG:track' and '/' in val:
                        val = val[:val.index('/')]
                    if key == 'TAG:date' and '-' in val:
                        val = val[:val.index('-')]
                    for i in ['utf-8', 'cp1252']:    
                        try:
                            val = val.decode(i)
                            break
                        except UnicodeEncodeError:
                            pass
                        except UnicodeDecodeError:
                            pass
                    try:
                        self._info[TAG_TRANSLATION[key]] = self.TAG_TYPES[TAG_TRANSLATION[key]](val)
                    except ValueError:
                        if val != '':
                            self.logit_error(logger, "Interpretation failed for %s: key (%s) value (%s)", self.filename, key, val)

    def __str__(self):
        try:
            return subprocess.check_output(self._avprobe_command())
        except OSError:
            return subprocess.check_output(self._ffprobe_command())


class audio_info_cached(audio_info):
    LOG_PREFIX = 'AudInfoCa:'

    """Class to hold and handle information of an audio file. See also parent class :py:class:`audio_info`.
    This class caches the information to have a faster access.

    :param str filename: path to the audio file
    :param str cache_filename: path to the cache file
    :param logger: Optionally logger instance to use instead of creating one

    **Example:**

    .. code-block:: python

        a = audio_info_cached(os.path.join(basepath, 'example', 'audio.mp3'), os.path.join(basepath, 'example', 'audio.json'))
        for key in a.keys():
            print key, type(a.get(key)), a.get(key)

    Will result to the following output:

    .. code-block:: text

        album <type 'str'> A/B
        title <type 'str'> No Good
        track <type 'int'> 1
        artist <type 'str'> Kaleo
        genre <type 'str'> Rock
        year <type 'int'> 2016
        duration <type 'float'> 236.094694
    """

    def __init__(self, filename, cache_filename, load_all_on_init=True, callback_on_data_storage=None):
        audio_info.__init__(self, filename)
        self._cached_data = caching.property_cache_json(audio_info(filename), cache_filename, load_all_on_init, callback_on_data_storage)

    def get(self, key, default=None, logger=None):
        """
        This gets an information of the video by a key.

        :param key: The key (name) of the information to get
        :param default: The default value to be returned, if no information with that key exists
        :returns: The information for the given key
        """
        self.logit_debug(logger, "Property request (%s) for %s", key, os.path.basename(self.filename))
        return self._cached_data.get(key, default, logger=logger)
